require 'middleclass'
require 'vector'
require 'AnAL'

Actor = class('Actor')

Pony = class('Pony',Actor)
function Pony:initialize(x,y,size)
    self.id = global_increment_id()
    self.orientation = 1 -- look direction
    self.health = 1
	self.angle = 0
    self.controlled = false -- by player
    self.stand_on = false -- on ground
    self.touching = false -- object
    self.direction = 0 --movement direction flag (0 - no move, -1 - left, 1 - right)
    self.jump = false -- jump flag (0 - no jump, 1 - jump)

	self.size = size or 40
    --physics
    self.position = vector(x,y)
    self.body = love.physics.newBody(world, x, y)
    self.shape = love.physics.newCircleShape(self.body,0,0,self.size) --main shape
	self.shape:setData(self)
    self.shape:setCategory(ACTOR)
	--self.shape:setMask(ACTOR)
    self.shape:setFriction(0.5)
	self.shape:setRestitution(0)
    self.body:setMassFromShapes()
	local mass = self.body:getMass()
	self.body:setMass(0,0,mass,0)
    self.speed = 50*mass -- normal acceleration
    self.speed_factor = 1 -- speed coefficient for run (1 - normal walk)
    self.max_speed = self.speed -- speed cap
    self.jump_impulse = -5*mass -- jump impulse
    self.jump_factor = 2 -- jump impulse modificator (stronger jump after crouch)
    self.jump_force = -self.jump_impulse -- in-jump control force

end

function Pony:dump()
	print(self.nx, self.ny, self.angle)
end

function Pony:setAnimations(list,default)
	self.animations = {}
	for index,anim in pairs(list) do
		img = love.graphics.newImage(anim.file)
		self.animations[index] = newAnimation(img, anim.size, anim.size, anim.delay, anim.length)
		self.animations[index]:setMode("loop")
	end
	self.animation = self.animations[default]
end

function Pony:preremove()
	if DEBUG then print("Prepare for remove",self, self.id) end
    self.shape:setMask(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
    self.preremoved = true
end
function Pony:destroy()
	--TODO: add nice particle effect for disappear
    if self.preremoved then
		if DEBUG then print("Removing",self, self.id) end
        self.shape:destroy()
        self.body:destroy()
    end
end

function Pony:draw()
    local x,y = self.position.x, self.position.y
	if self.controlled then love.graphics.setColor(255,255,255) else love.graphics.setColor(127,127,127) end
	self.animation:draw(x+self.orientation*4,y+10,self.angle,self.orientation, 1, 64,64)
    if DEBUG then
		love.graphics.circle("line", x,y,self.size)
		if self.stand_on then
			love.graphics.setColor(255,0,0)
			love.graphics.circle("fill", self.cx,self.cy,4)
			love.graphics.line(self.cx, self.cy, self.cx+self.nx,self.cy+self.ny)
		end
		if self.touching then
			love.graphics.setColor(0,0,255)
			love.graphics.circle("fill", self.tx,self.ty,4)
		end
    end
	--TODO: draw in framebuffer
end

function Pony:stop()
	self.direction = 0
	self.jump = false
	self.jump_factor = 2
	self.speed_factor = 1
end

function Pony:left(v)
	if v then self.direction = -1
	elseif self.direction == -1 then self.direction = 0 end
end

function Pony:right(v)
	if v then self.direction = 1
	elseif self.direction == 1 then self.direction = 0 end
end

function Pony:up(v)
	if v then self.jump = true
	else self.jump = false end
end

function Pony:down(v)
	if v then
		self.jump_factor = self.jump_factor * 2
		self.speed_factor = self.speed_factor / 2
	else
		self.jump_factor = self.jump_factor / 2
		self.speed_factor = self.speed_factor * 2
	end
end

function Pony:grab()
	if self.grab_joint then self.grab_joint:destroy(); self.grab_joint = false; self.grabbed = nil
	elseif self.touching then
		--TODO: check touching object state. very heavy, dangerous etc.
		self.grabbed = self.touching
		self.grabbed.body:setPosition(self.position.x, self.position.y-10)
		self.grab_joint = love.physics.newRevoluteJoint(
			self.body, self.grabbed.body,
			self.position.x, self.position.y-10
		)
		self.grab_joint:setMotorEnabled(true)
		self.grab_joint:setMaxMotorTorque(10)
	end
end

function Pony:action()
	if self.grabbed and self.grabbed.activate then
		self.grabbed.activate()
	end
end

function Pony:update(dt)
    self.body:wakeUp()
    self.position.x, self.position.y = self.body:getPosition()
    local vx, vy = self.body:getLinearVelocity()

    if self.controlled then
        -- movement
        if self.stand_on then --on ground
            if self.direction ~= 0 then
                if math.abs(vx) < self.max_speed*self.speed_factor then --speed cap
                    self.body:applyForce(self.direction*self.speed*self.speed_factor,0)
                end
				self.orientation = self.direction
			end
            if self.jump then
				--local impulse = self.jump_impulse*self.jump_factor
                --self.body:applyImpulse(self.nx*impulse,self.ny*impulse)
				self.body:applyImpulse(self.nx*self.jump_factor,self.ny*self.jump_factor)
                --and apply Newton third law
                self.stand_on.body:applyImpulse(-self.nx,-self.ny,self.cx,self.cy)
                --TODO: play jump sound
                self.jump = false
            end
--~             if (self.dash == 1) and (self.dash_cooldown <= 0) then
--~                 self.body:applyImpulse(-self.jump_impulse*self.direction,0)
--~                 self.dash = 0
--~                 self.dash_cooldown = 3 --sec
--~             end
        else --in air control
            if self.direction ~= 0 then
                force = self.jump_force*self.direction
                self.body:applyForce(self.jump_force*self.direction,0)
				self.orientation = self.direction
            end
        end
    end
	if self.stand_on then
		if math.abs(vx) > 10 then
			self.animation = self.animations.walk
			self.animation:setSpeed(math.abs(vx/100))
			--TODO: play step sound with rate according to walk speed
		else
			self.animation = self.animations.stand
			self.animation:setSpeed(0.1)
		end
	else
		-- TODO: set animation state to jump
		-- and update it according to vertical velocity
		if math.abs(vy) > 10 then
			self.animation = self.animations.stand --placeholder
			self.animation:setSpeed(1)
		end
	end
	self.animation:update(dt)
	--update attached body (aim gun) (wtf, srsly)
	if self.grab_joint then
		local mx,my = camera:mousepos():unpack()
		local angle = math.atan2(
			self.grabbed.position.y-my,
			self.grabbed.position.x-mx
		)
		self.grabbed.body:setAngle(angle)
	end
end

--TODO: add collide callback
--play collide sounds and animations
--if collide speed too high, add damage
--function Pony:cb_collide(object, contact)
    --local cx,cy = contact:getPosition()
    --if instanceOf(StaticPlatform, object) or instanceOf(DynamicBody, object) then
        --if cy > self.y and (cx > self.x-9) and (cx < self.x + 9) then
        --TODO: play foot hit sound
        --else
        --TODO: play hit sound
        --end
    --end
--end
function Pony:cb_persist(object,contact)
    local cx,cy = contact:getPosition()
	local x,y = self.position:unpack()
    if cy > y and (cx >= x-(self.size-10)) and (cx <= x + (self.size-10)) then
        if instanceOf(Platform, object) or instanceOf(DynamicBody, object) or instanceOf(Actor, object) then
            self.stand_on = object
			self.cx = cx --contact position
			self.cy = cy
			self.nx,self.ny = contact:getNormal() --contact normal
			--if self.ny > 0 then self.ny = -self.ny end --wtf?
			self.angle = math.atan2(self.nx,-self.ny)
        end
	end
    if (self.orientation*(x-cx) < -10) then --touching at front
		if instanceOf(DynamicBody, object) or instanceOf(Actor, object) then
			self.touching = object
			self.tx = cx
			self.ty = cy
		end
	end
end
function Pony:cb_remove(object,contact)
    if instanceOf(Platform, object) or instanceOf(DynamicBody, object) or instanceOf(Actor, object) then
        self.stand_on = false
		self.touching = false
		self.angle = 0
    end
end
